import shutil
import tempfile
import weakref
from html import escape
from textwrap import dedent

import panel as pn

from .grid import Grid

__version__ = '0.0.3'

CSS_FIX = """
html { font-family: Roboto, sans-serif; }
#sidebar { background-color: #fafafa; }
#sidebar > .mdc-drawer__content > .mdc-list { box-sizing: border-box; height: 100%; }
#pn-Modal.mdc-dialog--open { background: rgba(0, 0, 0, 0.25); backdrop-filter: blur(2px); }
#pn-Modal .mdc-dialog__content { padding: 0; --mnn-debug-accordion-font: 0.75em monospace; }
#pn-Modal .mdc-dialog__surface { max-width: 75vw; }
:host(.card-title) h3 { font: var(--mnn-debug-accordion-font, inherit); }
"""

SIDEBAR_OUTER_WIDTH = 400
SIDEBAR_INNER_WIDTH = SIDEBAR_OUTER_WIDTH - 10 - 1

pn.extension(
    'terminal', 'gridstack', 'tabulator', 'plotly', 'mathjax', 'codeeditor',
    raw_css=[CSS_FIX], sizing_mode='stretch_width',
    notifications=True, design='material')


class Manganite:
    _nb_instance = None
    _server_instances = {}


    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', None) or 'Manganite App'
        description = kwargs.pop('description', None)

        if pn.state.curdoc: # shared environment
            Manganite._server_instances[pn.state.curdoc] = self
        else: # running in JupyterLab
            Manganite._nb_instance = self

        self._upload_dir = tempfile.mkdtemp(prefix='mnn_uploads__')
        self._finalizer = weakref.finalize(self, shutil.rmtree, self._upload_dir)

        self._init_terminal()
        self._init_debugger()

        self._layout = {'Description': pn.Column()}

        if description is not None:
            self._layout['Description'].append(pn.pane.Markdown(description))

        self._tabs = pn.Tabs(('Description', self._layout['Description']))
        
        self._header = pn.FlexBox(justify_content='end')
        self._header.append(self._debugger_button)

        self._sidebar = pn.FlexBox(
            flex_direction='column',
            flex_wrap='nowrap',
            width=SIDEBAR_INNER_WIDTH) # explicit width for proper initial terminal size
        self._sidebar.append('## Log')
        self._sidebar.append(self._optimizer_terminal)

        self._modal = pn.Column(
            '### Exceptions',
            pn.pane.Alert(
                dedent("""\
                    **Caution**: except for SyntaxErrors, cell previews show the code
                    **after it has been transformed by Manganite**, so it might
                    differ slightly from the original notebook."""),
                alert_type='warning',
                margin=(0, 10, 10, 10),
                stylesheets=['p { margin: 0; }']),
            self._exceptions,
            margin=(0, 0),
            stylesheets=[':host { width: 75vw; max-width: 150ch; }'])

        self._template = pn.template.MaterialTemplate(
            collapsed_sidebar=True,
            header=[self._header],
            header_background='#000228',
            sidebar=[self._sidebar],
            main=[self._tabs],
            sidebar_width=SIDEBAR_OUTER_WIDTH,
            modal=[self._modal],
            title=title
        ).servable()


    def _init_terminal(self):
        terminal_options = {'fontSize': 14}
        theme = pn.state.session_args.get('theme')
        if theme is None or theme[0] != 'dark':
            terminal_options['theme'] = {
                'background': '#fafafa',
                'foreground': '#000'
            }

        self._optimizer_terminal = pn.widgets.Terminal(
            write_to_console=True,
            options=terminal_options,
            stylesheets=['.terminal-container { width: 100% !important; } .xterm .xterm-viewport { width: auto !important; }'])
    

    def _init_debugger(self):
        self._exceptions = pn.Accordion(
            margin=(0, 0),
            stylesheets=['.accordion { margin: 0; }'])

        self._debugger_button = pn.widgets.Button(
            button_type='danger',
            name='Debug',
            icon='bug-off',
            visible=False,
            stylesheets=[':host { width: fit-content; order: 1; } .bk-TablerIcon { vertical-align: initial; }'])
        self._debugger_button.on_click(lambda e: self._template.open_modal())
    
    def get_tab(self, name):
        if name not in self._layout:
            self._layout[name] = Grid()
            self._tabs.append((name, self._layout[name]))
        
        return self._layout[name]
    

    def get_header(self):
        return self._header
    

    def get_upload_dir(self):
        return self._upload_dir


    def add_exception(self, cell_number, line_number, cell_source, error_class, error_message):
        if line_number is None:
            location = 'cell {}, magic command'.format(cell_number)
            line_number = 1
        else:
            location = 'cell {}, line {}'.format(cell_number, line_number)
        preview_title = '{} in {}: {}'.format(error_class, location, escape(error_message))
        notification_content = '{}<br><small>{}</small>'.format(error_class, location)

        preview = pn.widgets.CodeEditor(
            name=preview_title,
            value=cell_source,
            language='python' if error_class != 'UsageError' else 'sh',
            readonly=True,
            theme='github',
            sizing_mode='stretch_width',
            margin=(0, 0))

        # annotations seem to work only when set
        # after the widget has been added to the layout
        def annotate_line(e):
            preview.annotations = [{
                'row': line_number - 1,
                'column': 0,
                'text': error_message,
                'type': 'error'
            }]
        self._debugger_button.on_click(annotate_line)

        self._exceptions.append(preview)
        if len(self._exceptions) == 1:
            self._debugger_button.name = '1 exception'
            self._debugger_button.visible = True
            self._modal.insert(2,
                pn.pane.Alert(
                    dedent("""\
                        **Hint:** further exceptions below cell {}
                        might occur due to its incomplete execution.""".format(cell_number)),
                    margin=(0, 10, 10, 10),
                    stylesheets=['p { margin: 0; }']))
        else:
            self._debugger_button.name = '{} exceptions'.format(len(self._exceptions))

        if pn.state.loaded:
            pn.state.notifications.error(notification_content)
        else:
            pn.state.onload(lambda: pn.state.notifications.error(notification_content))


    @classmethod
    def get_instance(cls):
        if pn.state.curdoc:
            if pn.state.curdoc not in cls._server_instances:
                if cls._nb_instance and pn.state.curdoc is cls._nb_instance._template.server_doc():
                    cls._server_instances[pn.state.curdoc] = cls._nb_instance
            return cls._server_instances[pn.state.curdoc]
        return cls._nb_instance


def init(*args, **kwargs):
    mnn = Manganite(*args, **kwargs)
    return mnn


def load_ipython_extension(ipython):
    from .magics import ManganiteMagics
    init()
    ipython.register_magics(ManganiteMagics)


def _jupyter_server_extension_points():
    return [{
        'module': 'manganite.jupyter'
    }]



import argparse
import sys

from panel import __version__ as pn_version
from panel.command.serve import Serve as PnServe

from manganite import __version__, preprocessor


def main():
    parser = argparse.ArgumentParser(prog='mnn')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {} (panel {})'.format(__version__, pn_version))

    subs = parser.add_subparsers(title='subcommands')

    serve_subparser = subs.add_parser(PnServe.name, help=PnServe.help)
    serve_subcommand = PnServe(parser=serve_subparser)
    serve_subparser.set_defaults(invoke=serve_subcommand.invoke)

    if len(sys.argv) == 1:
        args = parser.parse_args(['--help'])
        args.invoke(args)
        sys.exit()

    args = parser.parse_args()
    preprocessor._patch_python_exporter()

    args.invoke(args)

if __name__ == '__main__':
    main()



import ast
import re
import sys
import traceback
from collections import namedtuple
from datetime import date, datetime
from shlex import split

import ast_scope
import panel as pn
import param
from pandas import DataFrame
from IPython.core.magic_arguments import MagicArgumentParser
from IPython.core.error import UsageError

from manganite import Manganite
from .file_picker import FilePicker


class BoolWrapper(param.Parameterized):
    value = param.Boolean()


class NumberWrapper(param.Parameterized):
    value = param.Number()


class StringWrapper(param.Parameterized):
    value = param.String()


class DatetimeWrapper(param.Parameterized):
    value = param.Date()


class DataFrameWrapper(param.Parameterized):
    value = param.DataFrame()


WrappableTypes = {
    bool: BoolWrapper,
    int: NumberWrapper,
    float: NumberWrapper,
    str: StringWrapper,
    date: DatetimeWrapper,
    datetime: DatetimeWrapper,
    DataFrame: DataFrameWrapper}


def inspect_var(ns: dict, name: str):
    assert name.isidentifier()

    if name not in ns:
        return 'undefined'
    if isinstance(ns[name], param.Parameterized):
        return 'wrapped'
    if isinstance(ns[name], tuple(WrappableTypes.keys())):
        return 'wrappable'
    return 'non_wrappable'


class CellTransformer(ast.NodeTransformer):
    def __init__(self, scope_info, ns):
        self.scope_info = scope_info
        self.ns = ns
        self.stores = set()
        self.loads = set()
        self.undef_stores = set()
        self.undef_loads = set()


    def visit_Name(self, node):
        if not isinstance(self.scope_info[node], ast_scope.scope.GlobalScope):
            return node
        
        var_state = inspect_var(self.ns, node.id)

        # collect potentially deferred references
        if var_state == 'undefined':
            if isinstance(node.ctx, ast.Store):
                self.undef_stores.add(node.id)
            else:
                self.undef_loads.add(node.id)

        if isinstance(node.ctx, ast.Store):
            self.stores.add(node.id)
        else:
            self.loads.add(node.id)
        
        if var_state != 'wrapped':
            return node

        return ast.Attribute(
            value=ast.Name(id=node.id, ctx=ast.Load()),
            attr='value',
            ctx=node.ctx)


CellTransformInfo = namedtuple('CellTransformInfo', ['source', 'stores', 'loads', 'new', 'undefined'])


class CellManager():
    def __init__(self, ns):
        self.ns = ns
        self.deferred = {}
        self.panels = {}
        self.process_callbacks = {}
        self.cell_count = 0


    def transform(self, source) -> CellTransformInfo:
        source_tree = ast.parse(source)
        scope_info = ast_scope.annotate(source_tree)

        transformer = CellTransformer(scope_info, self.ns)
        transformed_tree = transformer.visit(source_tree)
        ast.fix_missing_locations(transformed_tree)
        return CellTransformInfo(
            ast.unparse(transformed_tree),
            transformer.stores,
            transformer.loads,
            transformer.undef_stores,
            transformer.undef_loads - transformer.undef_stores)


    def wrap(self, name: str, widget_attrs=None):
        if widget_attrs is None:
            for t, Wrapper in WrappableTypes.items():
                if isinstance(self.ns[name], t):
                    self.ns[name] = Wrapper(name=name, value=self.ns[name])
                    break
        else:
            var_type = type(self.ns[name])
            widget_type = widget_attrs['type']
            params = widget_attrs['params']
            if var_type == int:
                if widget_type == 'slider' and type(params) == str:
                    constraints = list(map(lambda x: int(x), params.split(':')))
                    if 2 <= len(constraints) <= 3 and constraints[0] < constraints[1]:
                        start, end = constraints[:2]
                        step = constraints[2] if len(constraints) == 3 else 1
                        self.ns[name] = pn.widgets.IntSlider(
                            name=name,
                            value=self.ns[name],
                            start=start,
                            end=end,
                            step=step)
                else:
                    self.ns[name] = pn.widgets.IntInput(name=name, value=self.ns[name])
            elif var_type == float:
                if widget_type == 'slider' and type(params) == str:
                    constraints = list(map(lambda x: float(x), params.split(':')))
                    if 2 <= len(constraints) <= 3 and constraints[0] < constraints[1]:
                        start, end = constraints[:2]
                        step = constraints[2] if len(constraints) == 3 else min(0.1, (end - start) / 10)
                        self.ns[name] = pn.widgets.FloatSlider(
                            name=name,
                            value=self.ns[name],
                            start=start,
                            end=end,
                            step=step)
                else:
                    self.ns[name] = pn.widgets.FloatInput(name=name, value=self.ns[name])
            elif var_type == str:
                if widget_type == 'select' and type(params) == str:
                    if params in self.ns and isinstance(self.ns[params], (list, set, tuple)):
                        options = list(self.ns[params])
                        value = self.ns[name] if self.ns[name] in options else options[0]
                        self.ns[name] = pn.widgets.Select(name=name, options=options, value=value)
                elif widget_type == 'radio' and type(params) == str:
                    if params in self.ns and isinstance(self.ns[params], (list, set, tuple)):
                        options = list(self.ns[params])
                        value = self.ns[name] if self.ns[name] in options else options[0]
                        self.ns[name] = pn.widgets.RadioBoxGroup(name=name, options=options, value=value)
                elif widget_type == 'text':
                    self.ns[name] = pn.widgets.TextInput(name=name,value=self.ns[name])
                elif widget_type == 'file':
                    accept = params if type(params) == str else None
                    self.ns[name] = FilePicker(name=name, accept=accept)
            elif var_type == bool:
                label = params if type(params) == str else name
                if widget_type == 'switch':
                    self.ns[name] = pn.widgets.Switch(
                        name=label, value=self.ns[name],
                        stylesheets=[':host { width: 40px; }'])
                else:
                    self.ns[name] = pn.widgets.Checkbox(name=label, value=self.ns[name])
            elif var_type == date:
                self.ns[name] = pn.widgets.DatePicker(name=name, value=self.ns[name])
            elif var_type == datetime:
                self.ns[name] = pn.widgets.DatetimePicker(name=name, value=self.ns[name])
            elif var_type == DataFrame:
                self.ns[name] = pn.widgets.Tabulator(self.ns[name])


    def process_exception(self, cell_number, cell_source, err: Exception):
        mnn = Manganite.get_instance()
        error_class = err.__class__.__name__
        error_message = err.args[0]

        if isinstance(err, SyntaxError):
            line_number = err.lineno
        elif isinstance(err, UsageError):
            line_number = None
        else:
            line_number = traceback.extract_tb(err.__traceback__)[1].lineno
        
        pn.state.log(err, level='error')
        mnn.add_exception(
            cell_number=cell_number,
            line_number=line_number,
            cell_source=cell_source,
            error_class=error_class,
            error_message=error_message)


    def add_cell(self, raw_source: str, process_var=None, widget_attrs=None):
        self.cell_count += 1
        cell_number = self.cell_count

        try:
            _, stores, loads, new, undefined = self.transform(raw_source)
        except Exception as err:
            self.process_exception(cell_number, raw_source, err)
            return

        defer = process_var is not None
        first_run = True
        def run_cell(*events):
            nonlocal first_run

            source = self.transform(raw_source).source
            try:
                exec(source, self.ns, self.ns)
            except Exception as err:
                self.process_exception(cell_number, source, err)
                return

            for name in stores:
                var_state = inspect_var(self.ns, name)

                if var_state == 'wrappable':
                    if widget_attrs and widget_attrs['name'] == name:
                        self.wrap(name, widget_attrs)
                    else:
                        self.wrap(name)
                
                if var_state != 'undefined':
                    if name in self.deferred and len(self.deferred[name]):
                        for cb in self.deferred[name]:
                            cb()
                        self.deferred.pop(name)

            if widget_attrs and widget_attrs['name'] in self.ns:
                name = widget_attrs['name']
                widget_attrs['display'](self.ns[name])

            if first_run:
                first_run = False
                if not process_var:
                    for name in loads - stores:
                        var_state = inspect_var(self.ns, name)
                        if var_state == 'wrapped':
                            # re-evaluate dependent cells only on actual change, except for file pickers,
                            # where the same name can refer to new file contents
                            onlychanged = not isinstance(self.ns[name], FilePicker)
                            self.ns[name].param.watch(run_cell, ['value'], onlychanged=onlychanged)

        deferred_deps = undefined & self.deferred.keys()
        process_deps = loads & self.process_callbacks.keys()

        if len(deferred_deps):
            defer = True
            for name in deferred_deps:
                self.deferred[name].add(run_cell)

        if len(process_deps):
            defer = True
            for name in process_deps:
                self.process_callbacks[name].add(run_cell)

        if defer:
            if process_var:
                self.process_callbacks[process_var] = set()
            for name in new:
                self.deferred[name] = set()
        else:
            run_cell()
        
        return run_cell
    

    def add_process_cell(self, args, raw_source):
        run_cell = self.add_cell(raw_source, process_var=args.returns)
        label = args.on[1]

        mnn = Manganite.get_instance()
        def run_process(*events):
            sys.stdout = mnn._optimizer_terminal
            sys.stderr = mnn._optimizer_terminal
            mnn._optimizer_terminal.write(
                '\033[32;1m[{}]\nExecuting "{}"...\033[0m\n\n'.format(
                    datetime.now().isoformat(sep=' ', timespec='seconds'),
                    label))
            try:
                run_cell()
                for cb in self.process_callbacks[args.returns]:
                    cb()
            finally:
                sys.stdout.flush()
                sys.stderr.flush()
                mnn._optimizer_terminal.write('\n\n')
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__
        
        button = pn.widgets.Button(
            name=label,
            stylesheets=[':host { width: fit-content; }'])
        button.on_click(run_process)
        if args.tab is not None:
            button.styles['grid_column_end'] = 'span 1'
            Manganite.get_instance().get_tab(args.tab).append(button)
        else:
            Manganite.get_instance().get_header().append(button)


    def add_widget_cell(self, args, raw_source):
        def display_widget(widget):
            if args.var in self.panels:
                self.panels[args.var].object = widget
            else:
                self.panels[args.var] = pn.panel(widget)
                tab_grid = Manganite.get_instance().get_tab(args.tab)
                grid_cell = pn.Column(
                    pn.pane.Markdown('### {}'.format(args.header or args.var)),
                    self.panels[args.var])

                y, x, w = args.position
                if y >= 0:
                    grid_cell.styles['grid_row_start'] = str(y + 1)
                if x >= 0 and x < 6:
                    grid_cell.styles['grid_column_start'] = str(x + 1)
                grid_cell.styles['grid_column_end'] = 'span {}'.format(w if 0 < w <= 6 else 3)

                tab_grid.append(grid_cell)

        widget = {
            'name': args.var,
            'type': args.type[0],
            'params': args.type[1] if len(args.type) > 1 else None,
            'display': display_widget}
        self.add_cell(raw_source, widget_attrs=widget)


    def add_magic_cell(self, arg_line, raw_source):
        parser = MagicArgumentParser()
        subparsers = parser.add_subparsers(dest='magic_type')

        process_parser = subparsers.add_parser('execute')
        process_parser.add_argument('--on', type=str, nargs=2, required=True)
        process_parser.add_argument('--tab', type=str, required=False)
        process_parser.add_argument('--returns', type=str, required=True)

        widget_parser = subparsers.add_parser('widget')
        widget_parser.add_argument('--var', type=str, required=True)
        widget_parser.add_argument('--tab', type=str, required=True)
        widget_parser.add_argument('--type', type=str, nargs='+', required=True)
        widget_parser.add_argument('--header', type=str, required=False)
        widget_parser.add_argument('--position', type=int, nargs=3,
            required=False, default=(-1, -1, 3))

        argv = split(arg_line)

        # check for slider range arguments, like `-50:50:5` or `0.0:1.0:0.01`
        # and prepend a space so that negative values are not parsed as --args
        range_spec = re.compile(':'.join(3 * [r'[+-]?(\d*\.)?\d+']))
        argv = [' ' + arg if range_spec.match(arg) else arg for arg in argv]

        try:
            args = parser.parse_args(argv)
        except UsageError as err:
            self.cell_count += 1
            self.process_exception(self.cell_count, '%%mnn ' + arg_line, err)
            return
        if args.magic_type == 'execute':
            self.add_process_cell(args, raw_source)
        elif args.magic_type == 'widget':
            self.add_widget_cell(args, raw_source)



from IPython.core.magic import Magics, magics_class, cell_magic, needs_local_scope


@magics_class
class ManganiteMagics(Magics):
    def __init__(self, *args, **kwargs):
        super(ManganiteMagics, self).__init__(*args, **kwargs)


    # dummy magic for running cells unmodified in Jupyter
    @needs_local_scope
    @cell_magic
    def mnn(self, arg_line, cell_source, local_ns):
        exec(cell_source, local_ns, local_ns)



from panel.layout.base import ListLike
from panel.reactive import ReactiveHTML


class Grid(ReactiveHTML, ListLike):
    # id attributes are altered by Panel
    # to make them unique for each instance
    _template = '''
    <div id="mnn-grid" class="mnn-grid">
      {% for obj in objects -%}
        <div id="mnn-grid__item" class="mnn-grid__item">
          ${obj}
        </div>
      {%- endfor %}
    </div>
    '''
    
    # `display: contents` on item wrappers
    # lets each item specify its own grid coordinates
    # by setting `grid-row-start` and `grid-column` on itself
    _stylesheets = ['''
    .mnn-grid {
      display: grid;
      grid-template-columns: repeat(6, 1fr);
      grid-template-rows: auto;
      place-items: stretch;
    }
                    
    .mnn-grid__item {
      display: contents;
    }
    ''']



def _load_jupyter_server_extension(app):
    import panel.io.jupyter_server_extension

    execution_template_lines = panel.io.jupyter_server_extension.EXECUTION_TEMPLATE.splitlines()
    i = next((i for i,line in enumerate(execution_template_lines) if 'PanelExecutor' in line), 0)
    execution_template_lines[i:i] = [
        'from manganite.preprocessor import _patch_python_exporter',
        '_patch_python_exporter()',
        ''
    ]

    panel.io.jupyter_server_extension.EXECUTION_TEMPLATE = '\n'.join(execution_template_lines)



import os

import panel as pn
import param
from panel.viewable import Viewer
from pathvalidate import sanitize_filename

from manganite import Manganite


class FilePicker(Viewer):
    value = param.FileSelector(label='Selected file')

    def __init__(self, accept=None, **params):
        self._create_subdir(params.get('name', ''))
        self._input = pn.widgets.FileInput(accept=accept)
        # call parent constructor only after initializing widgets
        # for `@param.depends` to work properly
        super().__init__(**params)
        self.param.value.path = os.path.join(self._path, '*')

        self._layout = pn.Column(self._input, self.param.value)


    def __panel__(self):
        return self._layout
    

    # this directory is deleted with its parent
    # on destruction of the current Manganite instance
    def _create_subdir(self, name):
        upload_dir = Manganite.get_instance().get_upload_dir()
        self._path = os.path.join(upload_dir, name)
        os.mkdir(self._path)


    # param==1.13.0 does not update `objects` properly on `path` change
    # so we need to trigger `update()` manually
    @param.depends('value:path', watch=True)
    def _update_selector_objects(self):
        self.param.value.update()


    @param.depends('_input.value', watch=True)
    def _save_upload(self):
        if self._input.value is not None:
            filename = sanitize_filename(self._input.filename)
            filepath = os.path.join(self._path, filename)
            self._input.save(filepath)

            # trigger value change on the first upload
            # or a re-upload of the currently selected file
            self.param.value.update()
            if len(self.param.value.objects) == 1 or self.value == filepath:
                self.value = filepath



import logging
import re
from textwrap import dedent
from uuid import uuid4

import nbconvert.exporters
import nbconvert.preprocessors
from IPython.core.inputtransformer2 import TransformerManager


class TransformManganiteMagicsPreprocessor(nbconvert.preprocessors.Preprocessor):
    _magic_pattern = re.compile(r'^\s*%%mnn\s+(.*)')
    _title_pattern = re.compile(r'^#\s*(.+)')
    _transformer = TransformerManager()


    def __call__(self, nb, resources):
        return self.preprocess(nb, resources)


    def preprocess(self, nb, resources):
        if not self.has_import(nb):
            return nb, resources

        description = '\n\n'.join([cell['source'] for cell in nb.cells if self.is_description_cell(cell)])
        title = self._title_pattern.match(description)

        nb.cells.insert(0, {
            'id': str(uuid4()),
            'cell_type': 'code',
            'outputs': [],
            'execution_count': 1,
            'metadata': {'tags': ['mnn-ignore']},
            'source': dedent("""\
                import manganite as _mnn_import
                from manganite.cell_manager import CellManager
                _mnn_import.init(title={0!r}, description={1!r})
                _mnn_cell_mgr = CellManager(globals())""".format(title.group(1) if title else None, description))
        })

        return super().preprocess(nb, resources)


    def preprocess_cell(self, cell, resources, index):
        if '%load_ext manganite' in cell['source']:
            cell['source'] = cell['source'].replace('%load_ext manganite', '')

        if 'mnn-ignore' in cell['metadata'].get('tags', []):
            return cell, resources

        if cell['cell_type'] == 'code':
            cell['source'] = self.transform_cell(cell['source'].lstrip())

        return cell, resources


    def has_import(self, nb):
        import_pattern = re.compile(r'import\s+manganite')
        cell_has_import = lambda cell: cell['cell_type'] == 'code' and import_pattern.match(cell['source'])

        return next((True for cell in nb.cells if cell_has_import(cell)), False)


    def is_description_cell(self, cell):
        if 'mnn-ignore' in cell['metadata'].get('tags', []):
            return False

        return cell['cell_type'] == 'markdown'
    

    def transform_cell(self, cell):
        lines = cell.splitlines()
        if len(lines):
            match = self._magic_pattern.match(lines[0])
            if match is None:
                cell = self.strip_system_calls(cell)
                return '_mnn_cell_mgr.add_cell({0!r})'.format(cell)
            else:
                cell = self.strip_system_calls('\n'.join(lines[1:]))
                return '_mnn_cell_mgr.add_magic_cell({0!r}, {1!r})'.format(match.group(1), cell)

        return cell


    # IPython's `!system` calls need to be stripped explicitly at this step
    # because this preprocessor converts cells into strings
    # so any parser later will ignore their contents
    def strip_system_calls(self, cell):
        cell = self._transformer.transform_cell(cell)
        # Bokeh's NotebookHandler follows the same pattern
        # with `magic` and `run_line_magic`
        cell = cell.replace('get_ipython().system', '')
        cell = cell.replace('get_ipython().getoutput', '')

        return cell


def _patch_python_exporter():
    log = logging.getLogger(__name__)
    log.warning('Patching nbconvert PythonExporter before Bokeh creates a NotebookHandler')

    cls = nbconvert.exporters.PythonExporter
    old_init = cls.__init__

    def new_init(self, config=None, **kw):
        old_init(self=self, config=config, **kw)
        self.register_preprocessor(TransformManganiteMagicsPreprocessor())

    cls.__init__ = new_init
